#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <time.h>
#include <crypto/cryptodev.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <libgen.h>
#include <sys/errno.h>
#include <pthread.h>
#include <netinet/sctp.h>
#include <netinet/sctp_uio.h>
#include <getopt.h>
#include <netinet/tcp.h>
#include <sys/stat.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>
#include <openssl/ssl.h>
#include <openssl/bio.h>
#include <openssl/err.h>

char *sendfilename = NULL;
int resetat=0;
char *cipher = NULL;
int cipher_num=-1;
int block_sf=0;
int direct_io_copy=0;
int force_sendfile=0;
int client_only=0;
int server_only=0;
BIO *bio_err = NULL;
int port_no = 9999;
long offset = 0;
in_addr_t remote = 0x100007f;
char *hdr = NULL;
char hdr_buf[1024];
int hdr_interval = 64*1024;
int hdr_size;

struct iovec hdr_iov[1];

static void
print_buffer(uint8_t *ptr, uint16_t len)
{
	int i;
	char buffer[64], str[4];
	int clear=0;
	memset(buffer, 0, sizeof(buffer));
	for(i=0; i<len; i++) {
		if ((ptr[i] < ' ') || (ptr[i] > '~')) {
			sprintf(str, ".");
		} else {
			sprintf(str, "%c", (char)ptr[i]);
		}
		strcat(buffer, str);
		clear += printf("%2.2x", ptr[i]);
		if ((i+1) % 16 == 0) {
			printf("\t\t%s\n", buffer);
			memset(buffer, 0, sizeof(buffer));
			clear = 0;
		} else {
			clear += printf(" ");
		}
	}
	if (clear) {
		printf("\t\t%s\n", buffer);
	}
}

static void 
destroy_ctx(SSL_CTX *ctx)
{
	SSL_CTX_free(ctx);
}


/*The password code is not thread safe*/
static int 
password_cb(char *buf, int num, int rwflag,void *userdata)
{
	if(num < strlen("password")+1)
		return(0);
	strcpy(buf,"password");
	return(strlen("password"));
}

void *
lis_thread(void *arg)
{
	int *lis_sd, li_fd, i, r, sf_ok;
	socklen_t slen;
	size_t siz;
	struct sockaddr_in sin;
	char foo;
	int sd, ret;
	off_t sbytes;
	char buffer[65536];
	char string[512];
	struct tls_record_layer *rlp;
	struct stat sbuf;
	int snd_file_fd=-1;
	SSL_CTX *ctx;
	int len;
	BIO *sbio, *ssl_bio, *io, *bio;
	SSL *ssl;
	DH *dhp;

	if (sendfilename != NULL) {
		if (direct_io_copy) {
			int src, dst;
			char bnt[128], dnt[128];
			char *bn, *dn;
			char buf[4096];
			ssize_t len;


			strncpy(bnt, sendfilename, sizeof(bnt));
			strncpy(dnt, sendfilename, sizeof(dnt));

			bn = basename(bnt);
			dn = dirname(dnt);

			if (dn == NULL) {
				printf("error: %s no dirname\n", sendfilename);
				exit(1);
			}

			src = open(bn, O_RDONLY);
			if (src < 0) {
				printf("cannot open %s for reading\n", bn);
				exit(1);
			}
			if (stat(bn, &sbuf)) {
				printf("Can't stat file %s err:%d\n",
				    bn, errno);
				exit(-1);
			}
			dst = open(sendfilename, O_WRONLY|O_CREAT|O_TRUNC|O_DIRECT);
			if (dst < 0) {
				printf("Can't open file %s err:%d\n",
				    sendfilename, errno);
				exit(-1);
			}
			len = sbuf.st_size;
			while (len > 0) {
				ssize_t out;
				ssize_t bytes = read(src, buf, sizeof(buf));
				if (bytes < 0) {
					printf("Can't read file %s err:%d\n",
					    bn,  errno);
					exit(-1);
				}
				out = write(dst, buf, bytes);
				if (out != bytes) {
					printf("Can't write file %s err:%d\n",
					    sendfilename,  errno);
					exit(-1);
				}
				len -= bytes;
			}
			close(src);
			close(dst);
		}


		if (stat(sendfilename, &sbuf)) {
			printf("Can't stat file %s err:%d\n",
			       sendfilename,
			       errno);
			exit(-1);
		}
		snd_file_fd = open(sendfilename, O_RDONLY);
		if (snd_file_fd == -1) {
			printf("Warning, can't do sendfile\n");
			printf("open failed err:%d filename:%s\n",
			       errno, sendfilename);
			sendfilename = NULL;
		}
	}
//	ctx = SSL_CTX_new(SSLv23_method());
	ctx = SSL_CTX_new(TLSv1_2_method());
	if (ctx == NULL) {
		printf("Can't create ctx err:%d\n", errno);
		exit(-1);
	}

	SSL_CTX_set_ecdh_auto(ctx, 1);
//	SSL_CTX_set_dh_auto(ctx, 1);

	if (!(SSL_CTX_use_certificate_chain_file(ctx, "server.pem"))) {
		printf("Can't read cert file server.pem\n");
		exit(-1);
	}
	SSL_CTX_set_default_passwd_cb(ctx, password_cb);
	if (!(SSL_CTX_use_PrivateKey_file(ctx,
					  "server.pem",
					  SSL_FILETYPE_PEM))) {
		printf("Can't read key file\n");
		exit(-1);
	}

	if(!(SSL_CTX_load_verify_locations(ctx, "root.pem", 0))) {
		printf("Can't read cert file root.pem\n");
		exit(-1);
	}

	if (cipher) {
		/* Add in the ciphers we want */
		if (SSL_CTX_set_cipher_list(ctx, cipher) == 0) {
			printf("No ciphers available when %s is set -- exiting\n",
			       cipher);
			exit(-1);
		}
		printf("set cipher list to %s\n", cipher);
	} else {
		printf("Not setting a cipher list -- all up in the air\n");
	}
	bio = BIO_new_file("dh1024.pem", "r");
	if (bio == NULL) {
		printf("Can't load DH file\n");
		exit(-1);
	}
	dhp = PEM_read_bio_DHparams(bio, NULL, NULL, NULL);
	BIO_free(bio);
	ret = SSL_CTX_set_tmp_dh(ctx, dhp);
	if (ret < 0) {
		printf("Can't set in DH params\n");
		exit(-1);
	}
	/* Lets get a connection */
	slen = sizeof(sin);
	lis_sd = (int *)arg;
	li_fd = *lis_sd;
	sd = accept(li_fd, (struct sockaddr *)&sin, &slen);
	if (sd == -1) {
		printf("Accept fails -- bye\n");
		exit(-1);
	}
	/* now receive one too */
	memset(buffer, 0, sizeof(buffer));
	ret = recv(sd, buffer, 1024, 0);
	if (ret != 5) {
		printf("Got %d not 5 '%s'\n", ret, buffer);
	}
	if (strncmp(buffer, "Hello", 5) != 0) {
		printf("Did not find a hello?\n");
		exit(-1);
	}
	/* Lets talk in the clear and tell him (with hello) to SSL */
	if (send(sd, "Hello", 5, 0) != 5) {
		printf("Error on send err:%d\n", errno);
		exit(-1);
	}

	printf("We are ready to SSL-a-cise things\n");
	/* Now setup TLS on the connection */
	sbio = BIO_new_socket(sd, BIO_NOCLOSE);
	if (sbio == NULL) {
		printf("%s:Can't setup bio err:%d\n",
		       __FUNCTION__,
		       errno);
		exit(-1);
	}
	ssl = SSL_new(ctx);
	if (ssl == NULL) {
		printf("%s:Can't setup ssl err:%d\n",
		       __FUNCTION__,
		       errno);
		exit(-1);
	}
//	SSL_set_bio(ssl, sbio, sbio);
//	if (SSL_accept(ssl) <= 0) {

	if (SSL_set_fd(ssl, sd) == 0) {
		printf("Can't set fd: err:%d\n", errno);
		ERR_print_errors(bio_err);
		exit(-1);
	}

	SSL_set_accept_state(ssl);
	if (SSL_do_handshake(ssl) <= 0) {
		printf("Can't do accept -- error err:%d\n", errno);
		ERR_print_errors(bio_err);
		exit(-1);
	}
	sf_ok = SSL_can_use_sendfile(ssl);
	if (sf_ok) {
		printf("We can use sendfile -- Yeah!\n");
	} else {
		printf("Sendfile is *not* possible\n");
		if (force_sendfile)
			exit(-1);
	}
	if (block_sf) {
		sf_ok = 0;
	}
	if (sf_ok && (snd_file_fd != -1)) {
		printf("Sending file %s over the SSL connection\n", sendfilename);
		sprintf(string, "SEND:%s:%d\n", sendfilename, (int)sbuf.st_size - (int)offset);
		len = strlen(string);
		r = SSL_write(ssl, string, len);
		if (r != len) {
			printf("Can't write out r:%d len:%d\n",
			       r, len);
			exit(-1);
		}
		r = BIO_flush(sbio);
		printf("Server sends:%s (%d bytes) -- ", string, len);
		printf("BIO_flush returns %d\n", r);
		if (hdr != NULL) {
			size_t nbytes, left;
			int ret;
			struct sf_hdtr sf_hdr;

			printf("sending hdr %s at interval %ld\n", 
			    hdr, (long) hdr_interval);
			sf_hdr.headers = hdr_iov;
			sf_hdr.hdr_cnt = 1;
			sf_hdr.trailers = NULL;
			sf_hdr.trl_cnt = 0;
			left = sbuf.st_size;
			while (offset < sbuf.st_size) {
				if (left > hdr_interval)
					nbytes = hdr_interval;
				else
					nbytes = left;
				if (hdr_iov[0].iov_base != NULL) {
					ret = sendfile(snd_file_fd, sd, (off_t)offset, nbytes,
					    &sf_hdr, &sbytes, SF_NOCACHE);
				} else {
					ret = SSL_write(ssl, hdr, hdr_size);
					if (ret != hdr_size) {
						printf("SSL write ret:%d\n", r);
						exit(-1);
					}
					ret = sendfile(snd_file_fd, sd, (off_t)offset, nbytes,
					    NULL, &sbytes, SF_NOCACHE);
				}
				if (ret < 0) {
					printf("Sendfile fails err:%d\n", errno);
					exit(-1);
				}
				offset += hdr_interval;
				left -= hdr_interval;

			}
		} else { 
			if (sendfile(snd_file_fd, sd, (off_t)offset, 0,
			     NULL, &sbytes, SF_NOCACHE)) {
				printf("Sendfile fails err:%d\n", errno);
				exit(-1);
			}
		}
		printf("Sendfile sends:%ld bytes filesize:%ld\n",
		       sbytes, sbuf.st_size);
	} else if (snd_file_fd != -1) {
		/* We do it the old fashion way by reading and writing */
		int sndlen, rr;
		printf("Sending file %s over the SSL connection the old fashioned way\n", sendfilename);
		sprintf(string, "SEND:%s:%d\n", sendfilename, (int)sbuf.st_size);
		len = strlen(string);
		r = SSL_write(ssl, string, len);
		if (r != len) {
			printf("Can't write out r:%d len:%d\n",
			       r, len);
			exit(-1);
		}
		r = BIO_flush(sbio);
		len = sbuf.st_size;
		while(len > 0) {
			if (len > sizeof(buffer)) {
				sndlen = sizeof(buffer);
			} else {
				sndlen = len;
			}
			if ((rr = read(snd_file_fd, buffer, sndlen)) > 0) {
				sndlen = rr;
				r = SSL_write(ssl, buffer, sndlen);
				if (r != sndlen) {
					printf("SSL write ret:%d\n", r);
					exit(-1);
				}
			} else {
				printf("Error read ret:%d not %d errno:%d\n", rr, sndlen, errno); 
				exit(-1);
			}
			len -= sndlen;
		}
	}
	r = SSL_write(ssl, "DONE\n", 5);
	if (r != 5) {
		printf("Can't write my DONEn ?? r:%d not 5\n",
		       r);
		exit(-1);
	}
	r = BIO_flush(sbio);	
	printf("BIO Flush ret:%d\n", r);
	memset(buffer, 0, sizeof(buffer));
	r = SSL_read(ssl, buffer, 3);
	if (r != 3) {
		printf("Only got %d while looking for bye\n", r);
	} else {
		struct timespec ts;
		printf("Sleep final 100 ms\n");
		ts.tv_sec = 0;
		ts.tv_nsec = 100000;
		nanosleep(&ts, NULL);
	}
	r = SSL_shutdown(ssl);
	if (!r) {
		printf("%s:server side first shutdown %d\n",
		       __FUNCTION__, r);
		shutdown(sd, 1);
		r = SSL_shutdown(ssl);
		printf("%s:Second shutdown ret:%d\n",
		       __FUNCTION__, r);
	}
	/* Close up please */
	printf("%s:Free ctx\n", __FUNCTION__);
	SSL_free(ssl);
	destroy_ctx(ctx);
	close(sd);
	if (snd_file_fd != -1) {
		close(snd_file_fd);
	}
out:
	return (NULL);
}

void
con_thread(int con_sd)
{
	struct timespec ts;
	struct sockaddr_in sin;
	socklen_t slen;
	uint16_t len;
	char buffer[65536];
	uint8_t output[1024];
	uint8_t tag[64];
	char fun[128];
	char filenameout[1024], *file=NULL, *num=NULL;
	FILE *out=NULL;
	BIO *sbio, *bio;
	int ret, fd, r, at, notdone, sz, rd, i;
	uint16_t nlen;
	SSL_CTX *ctx;
	SSL *ssl;
	DH *dhp;
	ssize_t read_size;


	/* Now Initialize the TLS stuff */
	ctx = SSL_CTX_new(TLSv1_2_method());
	if (ctx == NULL) {
		printf("Can't create ctx err:%d\n", errno);
		exit(-1);
	}
	if (!(SSL_CTX_use_certificate_chain_file(ctx, "client.pem"))) {
		printf("Can't read cert file server.pem\n");
		exit(-1);
	}
	SSL_CTX_set_default_passwd_cb(ctx, password_cb);
	if (!(SSL_CTX_use_PrivateKey_file(ctx,
					  "client.pem",
					  SSL_FILETYPE_PEM))) {
		printf("Can't read key file\n");
		exit(-1);
	}
	if(!(SSL_CTX_load_verify_locations(ctx, "root.pem", 0))) {
		printf("Can't read cert file root.pem\n");
		exit(-1);
	}
	if (cipher) {
		/* Add in the ciphers we want */
		if (SSL_CTX_set_cipher_list(ctx, cipher) == 0) {
			printf("No ciphers available when %s is set -- exiting\n",
			       cipher);
			exit(-1);
		}
		printf("set cipher list to %s\n", cipher);
	} else {
		printf("Not setting a cipher list\n");
	}
	bio = BIO_new_file("dhparam.pem", "r");
	if (bio == NULL) {
		printf("Can't load DH file\n");
		exit(-1);
	}
	dhp = PEM_read_bio_DHparams(bio, NULL, NULL, NULL);
	BIO_free(bio);
	ret = SSL_CTX_set_tmp_dh(ctx, dhp);
	if (ret < 0) {
		printf("Can't set in DH params\n");
		exit(-1);
	}

	/* setup the socket stuff */
	memset(&sin, 0, sizeof(sin));
	sin.sin_port = htons(port_no);
	sin.sin_family = AF_INET;
	sin.sin_len = sizeof(sin);
//	sin.sin_addr.s_addr = htonl(0x7f000001);
	sin.sin_addr.s_addr = remote;
	slen = sizeof(sin);
	ts.tv_sec = 0;
	ts.tv_nsec = 100000;
	/* Sleep a bit to let the server get its act together */
	printf("Sleep 100ms before connect\n");
	nanosleep(&ts, NULL);

	/* Now lets connect and talk in the clear */
	if (connect(con_sd, (struct sockaddr *)&sin, slen)) {
		printf("Connect fails to port err:%d\n", errno);
		exit(-1);
	}
	/* send a hello */
	if (send(con_sd, "Hello", 5, 0) != 5) {
		printf("Error on send err:%d\n", errno);
		exit(-1);
	}
	/* now receive one too */
	memset(buffer, 0, sizeof(buffer));
	ret = recv(con_sd, buffer, 1024, 0);
	if (ret != 5) {
		printf("Got %d not 5 '%s'\n", ret, buffer);
	}
	if (strncmp(buffer, "Hello", 5) != 0) {
		printf("Did not find a hello?\n");
		exit(-1);
	}
	/* Setup TLS */
	sbio = BIO_new_socket(con_sd, BIO_NOCLOSE);
	if (sbio == NULL) {
		printf("%s:Can't setup bio err:%d\n",
		       __FUNCTION__,
		       errno);
		exit(-1);
	}
	ssl = SSL_new(ctx);
	if (ssl == NULL) {
		printf("%s:Can't setup ssl err:%d\n",
		       __FUNCTION__,
		       errno);
		exit(-1);
	}
	SSL_set_bio(ssl, sbio, sbio);
	/* Now lets do the connect */
	printf("Doing SSL_connect on client side\n");
	if (SSL_connect(ssl) <= 0) {
		printf("Connection failed\n");
		exit(-1);
	}
	memset(buffer, 0, sizeof(buffer));
	r = SSL_read(ssl, buffer, 5);
	if (r < 5) {
		printf("Can't read control word\n");
		printf("r = %d, ssl_err = %d\n", r,
			SSL_get_error(ssl, r));
		exit(-1);
	}
	printf("Client reads 5 bytes:%s\n", buffer);
	if (strncmp(buffer, "DONE\n", 5) == 0) {
		/* We are done */
		printf("Got %s", buffer);
		goto done;
	}
	/* check if its SEND: */
	if (strcmp("SEND:", buffer) != 0) {
		printf("First bits *not* DONE and not SEND:\n");
		exit(-1);
	}
	/* 
	 * Ok to find out how much we are sending we need to
	 * read to a \n.
	 */
	at = 0;
	notdone = 1;
	memset(buffer, 0, sizeof(buffer));
	while (notdone) {
		r = SSL_read(ssl, &buffer[at], 1);
		if (r >= 0) {
			if (buffer[at] == '\n') {
				notdone = 0;
			}
			at++;
		}
		if (at >= sizeof(buffer)) {
			printf("No cr after %d bytes?\n", at);
			exit(-1);
		}
	}
	/* We now have filename:bytes\n in buffer */
	printf("We have the string %s\n", buffer);
	for(i=0; i<at; i++) {
		if (buffer[i] == ':') {
			file = buffer;
			num = &buffer[(i+1)];
			buffer[i] = 0;
			break;
		}
	}
	if ((file == NULL) || (num == NULL)) {
		printf("Could not find : separator in buf:%s\n",
		       buffer);
		exit(-1);
	}
	sprintf(filenameout,"%s.new", file);
	sz = strtol(num, NULL, 0);
	printf("Collecting %d bytes in new file:%s\n", sz, filenameout);
	out = fopen(filenameout, "w");
	if (out == NULL) {
		printf("Can't open %s err:%d\n", filenameout, errno);
		exit(-1);
	}
	
	if (hdr)
		read_size = 0;
	else
		read_size = sz;
	while (sz > 0) {

		if (read_size == 0 && hdr) {
			/* look for a header */
			r = SSL_read(ssl, buffer, hdr_size);
			if (r <= 0) {
				printf("Got a %d read return and still have %d to go eno:%d\n",
				    r, sz, errno);
				ERR_print_errors(bio_err);
				goto done;
			}
			if (!strncmp(hdr, buffer, hdr_size)) {
//				printf("found header %s\n", hdr);
			} else {
				buffer[strlen(hdr)] = '\0';
				printf("got %s, looking for header %s\n", buffer,hdr);
				exit(1);
			}
			read_size = hdr_interval;
		}
		if (sz > sizeof(buffer)) {
			rd = sizeof(buffer);
		} else {
			rd = sz;
		}
		if (rd > read_size)
			rd = read_size;

//		printf("%ld until header, reading %ld\n", read_size, (long)rd);

		if (resetat) {
			/* Do we reset yet? */
			if (sz >= resetat) {
				struct linger l;

				l.l_onoff = 1;
				l.l_linger = 0;
				r = setsockopt(con_sd, SOL_SOCKET, SO_LINGER, &l, sizeof(l));
				printf("Setting linger returns %d -- closing\n", r);
				close(con_sd);
				sleep(3);
			}
		}
		r = SSL_read(ssl, buffer, rd);
		if (r <= 0) {
			printf("Got a %d read return and still have %d to go eno:%d\n",
			       r, sz, errno);
			ERR_print_errors(bio_err);
			goto done;
		}
		sz -= r;
		if ((ret = fwrite(buffer, 1, r, out)) != r) {
			printf("write %d returns %d\n",
			       r, ret);
			goto done;
		}
		read_size -= r;

	}
	/* Ok we are all done */
	fflush(out);
	/* Should be DONE here */
	memset(buffer, 0, sizeof(buffer));
	r = SSL_read(ssl, buffer, 5);
	if (r != 5) {
		printf("Only got %d while looking for DONE\n", r);
		ERR_print_errors(bio_err);
		goto done;
	}
	if (strncmp(buffer, "DONE", 4) != 0) {
		printf("Can't find done.. got %s?\n", buffer);
		ERR_print_errors(bio_err);
		goto done;
	}
	printf("Sending the bye\n");
	r = SSL_write(ssl, "bye", 3);
	if (r != 3) {
		printf("Can't write my byte ?? r:%d not 3\n",
		       r);
		exit(-1);
	}
	r = BIO_flush(sbio);	
	printf("%s:Thats all folks!!\n", __FUNCTION__);
done:
	r = SSL_shutdown(ssl);
	printf("%s:Shutdown done\n", __FUNCTION__);
	if (r < 0) {
		printf("Shutdown failed ret:%d client side\n", r);
	}
	if (r == 0) {
		printf("%s:Calling shutdown again\n", __FUNCTION__);
		r = SSL_shutdown(ssl);
	}
	printf("%s:eFree ctx\n", __FUNCTION__);
	SSL_free(ssl);
	destroy_ctx(ctx);	
	/* close up we are done */
	close(con_sd);
	if (out) {
		fclose(out);
	}
}

static void
sigpipe_handler(int x)
{
}

static void
useage(const char *name)
{
      printf("Usage %s -p port [-f filename -c cipher -r resetamm -B ]\n",
	     name);
      printf("\nThe %s program will test the TLS-sendfile feature\n", name);
      printf("The program creates two threads, one a client the other a server.\n");
      printf("You pass it a minimum of the port (-p port) to specify to the\n");
      printf("server side what port to use. The server will listen on the port\n");
      printf("The client will connect to that port. They will exchange an initial\n");
      printf("'Hello' to synchronize each other and then they will setup connection in TLS.\n");
      printf("If no other argument is given they will then check the availablity of the\n");
      printf("ssl_sendfile command reporting if it is possible or not.\n");
      printf("If you pass in a -c cipher, the server will restrict itself to that OpenSSL cipher list\n");
      printf("If you pass in a filename, then if ssl_sendfile is possible the server will\n");
      printf("Do an ssl_sendfile with that file to the client. The client will create a\n");
      printf("new copy of that file in the local directory filename.new.\n");
      printf("Note that the sendfile happens all at once in this version\n");
      printf("If you give a -r resetamm the client will send a reset when it receives resetamm of data.\n");
      printf("The -B option can be supplied to block using sendfile\n");
}

int
main(int argc, char **argv)
{
	/* setup a listen socket. */
	int lsn_sd;
	int acc_sd;
	int con_sd;
	pthread_t t1, t2;
	socklen_t slen;
	int i;
	struct sockaddr_in sin;

	while((i = getopt(argc, argv, "r:o:p:f:c:I:n:H:h:i:BCDSF?")) != -1) {
		switch (i) {
		case 'B':
			block_sf = 1;
			break;
		case 'C':
			client_only = 1;
			break;
		case 'D':
			direct_io_copy = 1;
			break;
		case 'H':
			strncpy(hdr_buf, optarg, sizeof(hdr_buf));
			hdr = hdr_buf;
			printf("hdr set to %s\n", hdr);
			hdr_iov[0].iov_base = hdr;
			hdr_size = strlen(hdr) + 1;
			hdr_iov[0].iov_len = hdr_size;
			break;
		case 'h':
			strncpy(hdr_buf, optarg, sizeof(hdr_buf));
			hdr = hdr_buf;
			printf("hdr set to %s\n", hdr);
			hdr_size = strlen(hdr) + 1;
			break;
		case 'i':
			hdr_interval = strtol(optarg, NULL, 0);
			break;
		case 'S':
			server_only = 1;
			break;
		case 'F':
			force_sendfile = 1;
			break;
		case 'I':
			remote = inet_addr(optarg);
		case 'p':
			port_no = strtol(optarg, NULL, 0);
			break;
		case 'f':
			sendfilename = optarg;
			break;
		case 'c':
			cipher = optarg;
			break;
		case 'n':
			cipher_num = strtol(optarg, NULL, 0);
			break;
		case 'r':
			resetat = strtol(optarg, NULL, 0);
			break;
		case 'o':
			offset = strtol(optarg, NULL, 0);
			break;
		case '?':
		default:
		use:
			useage(argv[0]);
			exit(0);
			break;
		}
	}
	if ((port_no == 0) || (port_no > 65535)) {
		goto use;
	}
	if (!server_only) {
		con_sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if (con_sd == -1) {
			printf("Can't build connect socket err:%d\n",
		       errno);
			return(-1);
		}
	}
	/* Now lets bind the lsn_sd to port argv[1] */
	if (!client_only) {
		int foo;

		lsn_sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if (lsn_sd == -1) {
			printf("Can't build listen socket err:%d\n",
			    errno);
			return(-1);
		}
		if (setsockopt(lsn_sd, SOL_SOCKET, SO_REUSEADDR, &foo, sizeof(foo))) {
			printf("can't set SO_REUSEADDR\n");
			return(-1);
		}

		memset(&sin, 0, sizeof(sin));
		sin.sin_family = AF_INET;
		sin.sin_len = sizeof(sin);
		sin.sin_port = htons(port_no);
		if (bind(lsn_sd, (struct sockaddr *)&sin, sizeof(sin)) == -1) {
			printf("Can't bind port err:%d\n", errno);     
			return(-2);
		}
		slen = sizeof(sin);
		if (getsockname(lsn_sd, (struct sockaddr *)&sin, &slen) == -1) {
			printf("Can't get socket name.. err:%d\n", errno);
			return(-3);
		}
		if (ntohs(sin.sin_port) != port_no) {
			printf("Wanted port %d but got port %d -- goodbye\n",
			    port_no, ntohs(sin.sin_port));
			return(-4);
		}
		if (listen(lsn_sd, 1)) {
			printf("Can't listen on socket err:%d\n", errno);
			return(-5);
		}
	}
	/* Setup TLS stuff */
	SSL_library_init();
	OpenSSL_add_all_algorithms();
	SSL_load_error_strings();
	bio_err = BIO_new_fp(stderr, BIO_NOCLOSE);
	signal(SIGPIPE, sigpipe_handler);


#if 0
	if (!client_only && pthread_create(&t1, NULL, lis_thread, &lsn_sd)) {
		printf("Can't create thread1 err:%d\n", errno);
	}
#else
	if (!client_only) {
		pid_t r = fork();
		if (r != 0) {
			lis_thread(&lsn_sd);
			return(0);
		}
	}
#endif
	if (!server_only)
		con_thread(con_sd);
	return(0);
}
